/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: J2EEDomain.java 9398 2006-08-08 12:59:11 +0000 (Tue, 08 Aug 2006) durieuxp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.management.j2eemanagement;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.management.JMException;
import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.J2eeObjectName;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.management.cluster.BaseCluster;
import org.objectweb.jonas.management.monitoring.ClusterDaemonProxy;
import org.objectweb.jonas.management.monitoring.DomainMonitor;
import org.objectweb.jonas.management.monitoring.ServerProxy;
import org.objectweb.jonas.server.J2EEServer;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implements a managed object which represents either a management domain or a cluster.
 * This J2EEManagedObject is accessed mainly by jonasAdmin.
 * The complete interface is described in the following xml file :
 * @See src/org/objectweb/jonas/server/mbeans-descriptors.xml
 * @author Adriana Danes
 * @author Vivek Lakshmanan
 */
public class J2EEDomain extends J2EEManagedObject implements MBeanRegistration {

    /**
     * Management domain name
     */
    private String domainName = null;

    /**
     * J2EE Management logger
     */
    private static Logger logger = Log.getLogger("org.objectweb.jonas.management.j2eemanagement");

    /**
     * Ref to the J2EEServer of this server
     */
    private J2EEServer myJ2eeServer;

    private MBeanServer mbeanServer;

    /**
     * For slave servers, must keep here the connection
     * because there is no DomainMonitor.
     */
    private ServerProxy localServerProxy;

    /**
     * The associated DomainMonitor, in case this is the master node.
     */
    private DomainMonitor domainMonitor = null;

    private static J2EEDomain unique;

    /**
     * MBean constructor
     * @param objectName object name of the managed object
     * @throws JMException ..
     */
    public J2EEDomain(String objectName) throws JMException {
        super(objectName, false, false, false);
        ObjectName myObjectName = ObjectName.getInstance(objectName);
        domainName = myObjectName.getKeyProperty("name");
        unique = this;
    }

    /**
     * Set this server as the master
     */
    public void setMaster() {
        try {
            ObjectName dmOn = JonasObjectName.domainMonitor();
            domainMonitor = new DomainMonitor(domainName);
            if (mbeanServer != null) {
                mbeanServer.registerMBean(domainMonitor, dmOn);
            }
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Cannot create DomainMonitor", e);
        }
    }

    /**
     * Singleton: Each server (master or slave) must have 1 unique J2EEDomain object.
     * @return the unique instance of the J2EEDomain
     */
    public static J2EEDomain getInstance() {
        return unique;
    }

    /**
     * @param servername Name of the server
     * @return the connection to the MBeanServer of this server
     */
    public MBeanServerConnection getConnection(String servername) {
        if (servername.equals(myJ2eeServer.getServerName())) {
            // The local server
            logger.log(BasicLevel.DEBUG, "The local server");
            return localServerProxy.getConnection();
        } else {
            // Remote server: Only if we are in the master
            if (domainMonitor == null) {
                logger.log(BasicLevel.ERROR, "Cannot get Connection to remote server from a slave server");
                return null;
            }
            ServerProxy sp = domainMonitor.findServerProxy(servername);
            if (sp == null) {
                logger.log(BasicLevel.ERROR, "Unknown server: " + servername);
                return null;
            }
            return sp.getConnection();
        }
    }

    /**
     * Return the MBeans OBJECT_NAMEs of the servers belonging to this domain
     * @return The list of object names corresponding to the servers associated with this domain
     */
    public String[] getServers() {
        logger.log(BasicLevel.DEBUG, "");
        String[] sb = null;
        if (domainMonitor != null) {
            Collection col = domainMonitor.getServerList();
            sb = new String[col.size()];
            int i = 0;
            for (Iterator it = col.iterator(); it.hasNext(); ) {
                ServerProxy proxy = (ServerProxy) it.next();
                sb[i++] = proxy.getJ2eeObjectName();
            }
        } else {
            // Slave: Only itself
            sb = new String[1];
            sb[0] = myJ2eeServer.getObjectName();
        }
        return sb;
    }

    /**
     * Return the MBeans OBJECT_NAMEs of the servers belonging to this domain
     * but only if they are running.
     * @return The list of object names corresponding to the servers associated with this domain
     */
    public String[] getStartedServers() {
        logger.log(BasicLevel.DEBUG, "");
        ArrayList sb = new ArrayList();
        if (domainMonitor != null) {
            Collection col = domainMonitor.getServerList();
            for (Iterator it = col.iterator(); it.hasNext(); ) {
                ServerProxy proxy = (ServerProxy) it.next();
                if (proxy.getServerState() == ServerProxy.RUNNING) {
                    sb.add(proxy.getJ2eeObjectName());
                }
            }
        } else {
            // Slave: Only itself
            sb.add(myJ2eeServer.getObjectName());
        }
        return (String []) sb.toArray(new String [0]);
    }

    public String[] getClusterDaemons() {
        logger.log(BasicLevel.DEBUG, "");
        String[] sb = null;
        if (domainMonitor != null) {
            Collection col = domainMonitor.getClusterDaemonList();
            sb = new String[col.size()];
            int i = 0;
            for (Iterator it = col.iterator(); it.hasNext();) {
                ClusterDaemonProxy cdp = (ClusterDaemonProxy) it.next();
                sb[i++] = cdp.getObjectName();
            }
        } else {
            logger.log(BasicLevel.ERROR, "Operation allowed on master only");
        }
        return sb;
    }

    /**
     * Return the MBeans OBJECT_NAMEs of the clusters in this domain
     * @return The list of object names corresponding to the clusters in this domain
     */
    public String[] getClusters() {
        logger.log(BasicLevel.DEBUG, "");
        String[] sb = null;
        if (domainMonitor != null) {
            Collection col = domainMonitor.getTotalClusterList();
            sb = new String[col.size()];
            int i = 0;
            for (Iterator it = col.iterator(); it.hasNext();) {
                BaseCluster cl = (BaseCluster) it.next();
                if (cl == null) {
                    logger.log(BasicLevel.ERROR, "null cl elt");
                } else {
                    sb[i++] = cl.getObjectName();
                }
            }
        } else {
            logger.log(BasicLevel.ERROR, "Operation allowed on master only");
        }
        return sb;
    }

    /**
     * Get the domainMonitor, or null if slave server.
     * @return the domainMonitor, for the master server.
     */
    public DomainMonitor getDomainMonitor() {
        return domainMonitor;
    }

    /**
     * Create a logical cluster (by jonasAdmin)
     * @param name the cluster's name
     * @return the OBJECT_NAME corresponding to the created MBean
     */
    public String createCluster(String name) {
        logger.log(BasicLevel.DEBUG, "");
        if (domainMonitor != null) {
            BaseCluster cl = domainMonitor.createLogicalCluster(name);
            return cl.getObjectName();
        } else {
            logger.log(BasicLevel.ERROR, "Operation allowed on master only");
            return null;
        }
    }

    /**
     * Add the local J2EEServer
     * @param j2eeserver
     */
    public void addLocalServer(J2EEServer j2eeserver, Collection urls) {
        logger.log(BasicLevel.DEBUG, "");
        myJ2eeServer = j2eeserver;
        String svname = j2eeserver.getServerName();
        if (domainMonitor != null) {
            // Master case: must check if not already created.
            localServerProxy = domainMonitor.addLocalServer(svname, urls);
            // Read domain.xml after having added the local server
            domainMonitor.readDomainConfig();
        } else {
            // Slave case: create it in any case.
            localServerProxy = new ServerProxy(null, svname, urls, null);
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Adding server : "  + svname);
            }
            // give a name and register mbean
            ObjectName on;
            try {
                on = JonasObjectName.serverProxy(svname);
                localServerProxy.setObjectName(on.toString());
            } catch (MalformedObjectNameException e) {
                logger.log(BasicLevel.ERROR, "MalformedObjectName: "  + svname);
            }
        }
    }

    /**
     * Set the local server running and get its info
     * Cannot be done before because some MBeans must exist first.
     */
    public void setLocalServerRunning() {
        localServerProxy.getMonitoringInfo();
    }

    /**
     * Get a Cluster by its name
     * @param name The name of the cluster
     * @return the Cluster or null if unknown by the domain.
     */
    public BaseCluster findCluster(String name) {
        logger.log(BasicLevel.DEBUG, "");
        if (domainMonitor == null) {
            logger.log(BasicLevel.WARN, "discovery service not started");
            return null;
        }
        return domainMonitor.findCluster(name);
    }

    /**
     * Get a Server by its name
     * @param name The name of the server
     * @return the Server or null if unknown by the domain.
     */
    public J2EEServer findServer(String name) {
        logger.log(BasicLevel.DEBUG, "");
        if (domainMonitor == null) {
            logger.log(BasicLevel.WARN, "discovery service not started");
            return null;
        }
        return domainMonitor.findServer(name);
    }

    /**
     * Get a Server by its name
     * @param name The name of the server
     * @return the ServerProxy or null if unknown by the domain.
     */
    public ServerProxy findServerProxy(String name) {
        logger.log(BasicLevel.DEBUG, "");
        if (domainMonitor == null) {
            logger.log(BasicLevel.WARN, "discovery service not started");
            return null;
        }
        return domainMonitor.findServerProxy(name);
    }

    /**
     * @param server The MBean server in which the MBean will be registered.
     * @param name The object name of the MBean
     * @return The object name of the MBean
     */
    public ObjectName preRegister(MBeanServer mbeanServer, ObjectName name) throws Exception {
        logger.log(BasicLevel.DEBUG, "");
        if (name == null) {
            return J2eeObjectName.J2EEDomain(domainName);
        }
        this.mbeanServer = mbeanServer;
        String myName = getObjectName().toString();
        String argName = name.toString();
        if (myName.equals(argName)) {
            return name;
        } else {
            if (logger.isLoggable(BasicLevel.WARN)) {
                logger.log(BasicLevel.WARN, "Trying to register J2EEDomain MBean with the following name: "
                        + argName + " when the expected name is: " + myName);
            }
            return ObjectName.getInstance(domainName);
        }
    }

    /**
     * Add J2EEDomain MBean (myself) as listener to registration/unregistration notifications of
     * JOnAS management MBeans.
     * @param registrationDone Indicates whether or not the MBean has been successfully registered
     */
    public void postRegister(Boolean registrationDone) {
        logger.log(BasicLevel.DEBUG, "");
    }

    /**
     * @see javax.management.MBeanRegistration#preDeregister()
     */
    public void preDeregister() throws Exception {
        logger.log(BasicLevel.DEBUG, "");
    }

    /**
     * @see javax.management.MBeanRegistration#postDeregister()
     */
    public void postDeregister() {
        logger.log(BasicLevel.DEBUG, "");
    }



    /**
     * Start a Remote JOnAS Server
     * MBean implementation for jonasAdmin
     * @param serverName Name of the jonas server
     */
    public void startServer(String serverName) {
        if (domainMonitor == null) {
            logger.log(BasicLevel.ERROR, "Operation only possible on master");
            return;
        }
        ServerProxy proxy = domainMonitor.findServerProxy(serverName);
        proxy.startit();
    }

    /**
     * Stop a Remote JOnAS Server
     * MBean implementation for jonasAdmin
     * @param serverName Name of the jonas server
     */
    public void stopServer(String serverName) {
        if (domainMonitor == null) {
            logger.log(BasicLevel.ERROR, "Operation only possible on master");
            return;
        }
        ServerProxy proxy = domainMonitor.findServerProxy(serverName);
        proxy.stopit();
    }

    /**
     * @return Returns the description.
     */
    public String getDescription() {
        if (domainMonitor != null) {
            return domainMonitor.getDescription();
        } else {
            return "Unknown";
        }
    }

    /**
     * @return Returns the master.
     */
    public boolean isMaster() {
        return (domainMonitor != null);
    }

    /**
     * Return the domain name
     * @return domain name or cluster name
     */
    public String getMyName() {
        return domainName;
    }
}
